import re

from bleach import Cleaner
from html5lib.filters.base import Filter

# Tags, attributes, and styles allowed in Confluence storage format according to
# https://confluence.atlassian.com/doc/confluence-storage-format-790796544.html
ALLOWED_TAGS = ['a', 'ac:image', 'ac:layout', 'ac:layout-cell', 'ac:layout-section', 'ac:link',
                'ac:plain-text-link-body', 'ac:task-list', 'big', 'blockquote', 'br', 'code', 'div',
                'em', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'hr', 'i', 'li', 'ol', 'p', 'pre',
                'ri:attachment', 'ri:page', 'small', 'span', 'strong', 'sub', 'sup', 'table',
                'tbody', 'td', 'tfoot', 'th', 'thead', 'tr', 'tt', 'u', 'ul',
                'ac:structured-macro', 'ac:parameter',
                # Allow for full removal with RemovalFilter
                'style']

ALLOWED_ATTRS = {'*': ['class'], 'a': ['href', 'title'], 'span': ['style'],
                 'ri:page': ['ri:content-title'], 'ri:attachment': ['ri:filename'],
                 'ac:link': ['ac:anchor'],
                 'ri:url': ['ri:value'], 'ac:layout-section': ['ac:type'],
                 'ac:parameter': ['ac:name'],
                 'ac:structured-macro': ['ac:name', 'ac:schema-version'],
                 'td': ['rowspan', 'colspan'], 'th': ['rowspan', 'colspan']}

ALLOWED_STYLES = ['color', 'text-align', 'text-decoration']

# Tags that will be removed along with all of their descendants
REMOVED_TAGS = ['style']

EMPTY_TAG_REGEX = re.compile('<(hr|br)>')

class RemovalFilter(Filter):
    """Removes tags and all of their descendants."""
    def __iter__(self):
        stack = []
        for token in Filter.__iter__(self):
            if 'name' in token and token['name'] in REMOVED_TAGS:
                if token['type'] == 'StartTag':
                    stack.append(token['name'])
                elif token['type'] == 'EndTag':
                    stack.pop(-1)
            elif not stack:
                yield token


def sanitize_html(source):
    """Uses bleach to sanitize HTML of any tags and attributes that are
    invalid in Confluence storage format.

    Uses a regex to workaround https://github.com/mozilla/bleach/issues/28 in
    common cases.
    """
    html = Cleaner(
        tags=ALLOWED_TAGS,
        attributes=ALLOWED_ATTRS,
        styles=ALLOWED_STYLES,
        filters=[RemovalFilter],
        strip=True,
        strip_comments=True
    ).clean(source)
    return EMPTY_TAG_REGEX.sub(r'<\1/>', html)



"""Confluence page exporter that transforms notebook content into Confluence
XML storage format and posts it to an existing page.
"""
import os
import urllib.parse as urlparse

import requests

from .filter import sanitize_html
from .preprocessor import ConfluencePreprocessor
from nbconvert import HTMLExporter
from nbconvert.filters.markdown_mistune import IPythonRenderer, MarkdownWithMath
from traitlets import Bool, List, Unicode
from traitlets.config import Config


class ConfluenceExporter(HTMLExporter):
    """Converts a notebook into Confluence storage format XHTML and the
    notebook binary output cell assets into page attachments, and updates
    a given Confluence page with both.

    Attributes
    ----------
    server: str
        Base URL of the confluence server
    page_id: int
        Page ID to update
    notebook_filename: str
        Local filename of the notebook to be attached to the page

    url: traitlets.Unicode
        Human-readable Confluence page URL to convert to lookup page_id
    username: traitlets.Unicode
        Basic auth username
    password: traitlets.Unicode
        Basic auth password
    generate_toc: bool, optional
        Insert a Confluence table of contents macro at the top of the page (default: True)
    attach_ipynb: traitlets.Bool
        Attach the notebook ipynb to the page and link to it from the page footer (default: True)
    enable_style: traitlets.Bool
        Add the Jupyter base stylesheet to the page (default: True)
    enable_mathjax: traitlets.Bool
        Add MathJax to the page to render equations (default: False)
    """
    url = Unicode(config=True, help='Confluence URL to update with notebook content')
    username = Unicode(config=True, help='Confluence username')
    password = Unicode(config=True, help='Confluence password')
    generate_toc = Bool(config=True, default_value=True, help='Show a table of contents at the top of the page?')
    attach_ipynb = Bool(config=True, default_value=True, help='Attach the notebook ipynb to the page?')
    enable_style = Bool(config=True, default_value=True, help='Add basic Jupyter stylesheet?')
    enable_mathjax = Bool(config=True, default_value=False, help='Add MathJax to the page to render equations?')
    extra_labels = List(config=True, trait=Unicode(), help='List of additional labels to add to the page')

    @property
    def default_config(self):
        overrides = Config({
            'CSSHTMLHeaderPreprocessor': {
                'enabled': False
            },
            'HighlightMagicsPreprocessor': {
                'enabled': False
            },
            'TemplateExporter': {
                'exclude_input_prompt': True,
                'exclude_output_prompt': True
            },
            'TagRemovePreprocessor': {
                'remove_cell_tags': {'nocell', 'no-cell', 'no_cell'},
                'remove_input_tags': {'noinput', 'no-input', 'no_input'},
                'remove_all_outputs_tags': {'nooutput', 'no-output', 'no_output'},
                'enabled': True
            },
            'ExtractOutputPreprocessor': {
                'enabled': True
            },
        })

        c = super(ConfluenceExporter, self).default_config.copy()
        c.merge(overrides)
        return c

    def __init__(self, config, **kwargs):
        config.HTMLExporter.preprocessors = [ConfluencePreprocessor]
        config.HTMLExporter.filters = {
            'sanitize_html': sanitize_html,
        }

        super(ConfluenceExporter, self).__init__(config=config, **kwargs)
        self._preprocessors[-1].exporter = self

        self.template_path = [os.path.abspath(os.path.dirname(__file__))]
        self.template_file = 'confluence'
        # Must be at least a single character, or the header generator produces
        # an (invalid?) empty anchor tag that trips up bleach during
        # sanitization
        self.anchor_link_text = ' '

        self.server, self.page_id = self.get_server_info(self.url)
        self.notebook_filename = None

    def get_server_info(self, url):
        """Given a human visitable Confluence URL copy/pasted from the browser
        address bar, attempts to look up the programmatic page ID for use in
        working with the Confluence REST API.

        Parameters
        ----------
        url: str
            Human readable URL

        Returns
        -------
        2-tuple of str, int
            Confluence server base URL and programmatic page ID

        Raises
        ------
        Exception
            When the URL format is unknown, the Confluence API returns an error,
            or the Confluence API returns an unexpected result
        """
        pr = urlparse.urlparse(url)

        # Figure out the base URL of the server
        segs = pr.path.split('/')
        for i, seg in enumerate(segs):
            if seg in ('display', 'spaces'):
                server = pr.scheme + '://' + pr.netloc + '/'.join(segs[:i])
                break
        else:
            server = pr.scheme + '://' + pr.netloc

        # URL contains the pageId as a query arg
        # https://somewhere.com/pages/viewpage.action?pageId=123456
        query = urlparse.parse_qs(pr.query)
        if 'pageId' in query:
            # page ID is a query param
            return (server, int(query['pageId'][0]))

        # NOTE: Maybe these need to loop and find keys in the path, but I
        # don't have enough info yet on how these URLs can vary. Right now, just
        # using the path positions that I know about.

        # URL on Confluence Cloud contains the page ID in the path under a space
        # https://somewhere.atlassian.net/wiki/spaces/ASPACE/pages/123456/Page+Title
        if len(segs) > 5 and segs[4] == 'pages':
            page_id = int(segs[5])
            return (server, page_id)

        # URL on Confluence Server contains a space and page title requiring lookup
        # https://confluence.somewhere.com/display/ASPACE/Page+Title
        if len(segs) > 2 and segs[1] == 'display':
            # use space and page title to lookup the page ID
            space = segs[2]
            title = segs[3]

            resp = requests.get('{server}/rest/api/content?title={title}&spaceKey={space}'.format(server=server,
                                                                                                  title=title,
                                                                                                  space=space),
                                auth=(self.username, self.password)
                               )
            resp.raise_for_status()
            results = resp.json()['results']
            if not results:
                raise ValueError(
                    'Could not locate {} in {}. Ensure the page exists: '
                    'nbconflux will not create it for you.'.format(title, space)
                )
            return (server, int(resp.json()['results'][0]['id']))

        raise RuntimeError('Unknown URL format: ' + url)

    def update_page(self, page_id, body):
        """Updates the body of the page with new content.

        Parameters
        ----------
        page_id: int
            Confluence page ID
        body: str
            Confluence storage format content
            https://confluence.atlassian.com/doc/confluence-storage-format-790796544.html

        Raises
        ------
        Exception
            When Confluence API returns an error
        """
        # Fetch version number from the existing page so that we can increment it by 1.
        resp = requests.get('{server}/rest/api/content/{page_id}'.format(server=self.server,
                                                                         page_id=page_id),
                            auth=(self.username, self.password))
        resp.raise_for_status()
        content = resp.json()
        version = content['version']['number']
        # Newer Confluence requires title when posting the page back
        title = content['title']

        # Update the page with the new content.
        resp = requests.put('{server}/rest/api/content/{page_id}'.format(server=self.server,
                                                                         page_id=page_id),
                            json={
                               'version': {"number":version + 1},
                               'title': title,
                               'type': 'page',
                               'body': {
                                   'storage': {
                                       'representation': 'storage',
                                       'value': body
                                   }
                               }
                            },
                            auth=(self.username, self.password)
                           )
        resp.raise_for_status()

    def add_label(self, page_id, label):
        """Adds a label with global prefix to the page.

        Parameters
        ----------
        page_id: int
            Confluence page ID
        label: str
            Label to add

        Raises
        ------
        Exception
            When Confluence API returns an error
        """
        # Add the nbconflux label to the set of labels. OK if it already exists.
        resp = requests.post('{server}/rest/api/content/{page_id}/label'.format(server=self.server,
                                                                                page_id=page_id),
                             json=[dict(prefix='global', name=label)],
                             auth=(self.username, self.password))
        resp.raise_for_status()

    def add_or_update_attachment(self, filename, data, resources):
        """Creates or updates page attachments.

        Parameters
        ----------
        filename: str
            Local filename
        data: bytes
            Data to post
        resources: dict
            Additional nbconvert resources

        Returns
        -------
        request.Response
            Response from the Confluence server
        """
        basename = os.path.basename(filename)
        attachment = resources.get('attachments', {}).get(basename)
        if attachment is None:
            return
        files = {
            'file': (basename, data)
        }
        resp = requests.post(attachment.upload_url,
                             headers={
                                 'X-Atlassian-Token': 'nocheck'
                             },
                             files=files,
                             auth=(self.username, self.password))
        resp.raise_for_status()
        return resp

    def markdown2html(self, source):
        """Override the base class implementation to force empty tags to be
        XHTML compliant for compatibility with Confluence storage format.
        """
        renderer = IPythonRenderer(escape=False,
                                   use_xhtml=True,
                                   anchor_link_text=self.anchor_link_text)
        return MarkdownWithMath(renderer=renderer).render(source)

    def from_notebook_node(self, nb, resources=None, **kw):
        """Publishes a notebook to Confluence given a notebook object
        from nbformat.

        Parameters
        ----------
        nb: nbformat.notebooknode.NotebookNode
            Root of a notebook
        resources: dict
            Additional nbconvert resources

        Returns
        -------
        2-tuple
            Published Confluence storage format HTML and nbconvert resources
        """
        if self.notebook_filename is None:
            raise ValueError('only from_filename is supported')

        # Seed resources with option flags
        resources = resources if resources is not None else {}
        resources['generate_toc'] = self.generate_toc
        resources['enable_mathjax'] = self.enable_mathjax
        resources['enable_style'] = self.enable_style

        # Convert the notebook to Confluence storage format, which is XHTML-like
        html, resources = super(ConfluenceExporter, self).from_notebook_node(nb, resources, **kw)

        # Update the page with the new content
        self.update_page(self.page_id, html)
        # Add the nbconflux label to the page for tracking
        self.add_label(self.page_id, 'nbconflux')
        # If requested, add any extra labels to the page
        if self.extra_labels:
            for label in self.extra_labels:
                self.add_label(self.page_id, label)

        # Create or update all attachments on the page
        for filename, data in resources.get('outputs', {}).items():
            self.add_or_update_attachment(filename, data, resources)

        # Create or update the notebook document attachment on the page
        if self.attach_ipynb:
            with open(self.notebook_filename, encoding='utf-8') as f:
                self.add_or_update_attachment(self.notebook_filename, f.read(), resources)

        return html, resources

    def from_filename(self, filename, *args, **kwargs):
        """Publishes a notebook to Confluence given a local notebook filename.

        Parameters
        ----------
        filename: str
            Path to a local ipynb

        Returns
        -------
        2-tuple
            Published Confluence storage format HTML and nbconvert resources
        """
        # Preprocessor needs the filename to attach the notebook source file properly
        # so stash it here for later lookup
        self.notebook_filename = filename
        return super(ConfluenceExporter, self).from_filename(filename, *args, **kwargs)



"""Confluence page preprocessor that handles image and notebook
attachment versioning.
"""
import os

from collections import namedtuple

import requests

from nbconvert.preprocessors import Preprocessor
from traitlets import Instance, Any


Attachment = namedtuple('Attachment', 'id version download_url upload_url')


class ConfluencePreprocessor(Preprocessor):
    """Builds absolute URLs to versioned page attachments for use
    by HTMLExporter when rendering Confluence XHTML storage format.

    Attributes
    ----------
    exporter: ConfluenceExporter
        Exporter with information about the Confluence server and the page
        to be updated
    """
    exporter = Instance(klass='nbconflux.exporter.ConfluenceExporter', config=True)

    def preprocess(self, nb, resources):
        """Adds Attachment instances under resources['attachments']
        for every notebook output extracted by ExtractOutputPreprocessor
        and to be attached to the Confluence page.

        Also adds the filename of the notebook to be converted to a
        Confluence page under resources['notebook_filename'] and an associated
        Attachment for the notebook under resources['attachments'].

        Parameters
        ----------
        nb: nbformat.notebooknode.NotebookNode
            Root of a notebook
        resources: dict
            Additional nbconvert resources

        Returns
        -------
        2-tuple
            Modified nb and resources per the nbconvert Preprocessor API
            contract

        Note
        ----
        Uses the Confluence API to page through all attachments on the page in
        order to fetch their names and versions. This information is necessary
        to retain stable page-to-attachment version links in the page history.
        """
        # Get the attachments on the page, following ._links.next URLs until we know all attachment
        # names and versions.
        path = ('/rest/api/content/{page_id}/child/attachment?expand=version'
                .format(page_id=self.exporter.page_id))
        resources['attachments'] = {}
        while path:
            url = '{server}{path}'.format(server=self.exporter.server, path=path)
            resp = requests.get(url, auth=(self.exporter.username, self.exporter.password))
            resp.raise_for_status()
            attachments = resp.json()
            # Build a map from attachment filename to attachment ID and attachment version
            resources['attachments'].update(
                {result['title']: Attachment(result['id'], result['version']['number'], None, None)
                 for result in attachments['results']}
            )
            # Try to fetch the path (unfortunately, not full URL) of the next page of links
            path = attachments.get('_links', {}).get('next')

        # Notebook extreacted files to be attached to the page
        to_be_attached = dict(resources.get('outputs', {}))

        # consider the notebook itself an attachment that needs to be versioned
        if self.exporter.attach_ipynb:
            notebook_filename = os.path.basename(self.exporter.notebook_filename)
            to_be_attached[notebook_filename] = None
            resources['notebook_filename'] = notebook_filename

        for filename in to_be_attached:
            try:
                attachment_id, attachment_version, _, _ = resources['attachments'][filename]
            except KeyError:
                # The given filename is not yet an attachment in the page history so start at
                # version 0
                attachment_id = None
                attachment_version = 0
                upload_url = ('{server}/rest/api/content/{page_id}/child/attachment'
                              .format(server=self.exporter.server, page_id=self.exporter.page_id))
            else:
                upload_url = ('{server}/rest/api/content/{page_id}/child/attachment/{attachment_id}/data'
                              .format(server=self.exporter.server, page_id=self.exporter.page_id,
                                      attachment_id=attachment_id))

            # Populate the download url template
            download_url = ('{server}/download/attachments/{page_id}/{filename}?version={version}'
                            .format(server=self.exporter.server, page_id=self.exporter.page_id,
                                    filename=filename, version=attachment_version+1))

            # Keep the URL in the resources for later lookup in the page template
            resources['attachments'][filename] = Attachment(attachment_id, attachment_version, download_url, upload_url)

        return nb, resources


import getpass

from .exporter import ConfluenceExporter
from traitlets.config import Config


def notebook_to_page(notebook_file, confluence_url, username=None, password=None,
                     generate_toc=True, attach_ipynb=True, enable_style=True, enable_mathjax=False,
                     extra_labels=None):
    """Transforms the given notebook file into Confluence storage format and
    updates the given Confluence URL with its content.

    Attaches images to the page and links to pinned versions to preserve
    a complete page snapshot in the history. Also attaches a copy of the source
    notebook to the page and links to that version in the page footer.

    Parameters
    ----------
    notebook_file: str
        Relative or absolute path to the notebook to transform and post
    confluence_url: str
        Page URL to update with the notebook content. The page must
        already exist.
    username: str, optional
        Confluence username. Uses the current username if not specified.
    password: str, optional
        Confluence password. Prompts for the password if not given.
    generate_toc: bool, optional
        Insert a Confluence table of contents macro at the top of the page (default: True)
    attach_ipynb: bool, optional
        Attach the notebook ipynb to the page and link to it from the page footer (default: True)
    enable_style: bool, optional
        Include the Jupyter base stylesheet (default: True)
    enable_mathjax: bool, optional
        Include the MathJax script and configuration (default: False)
    extra_labels: list, optional
        Additional labels to add to the page (default: None)
    """
    if username is None:
        username = getpass.getuser()
    if password is None:
        password = getpass.getpass('Confluence password for {}:'.format(username))
    if extra_labels is None:
        extra_labels = []

    c = Config()
    c.ConfluenceExporter.url = confluence_url
    c.ConfluenceExporter.username = username
    c.ConfluenceExporter.password = password
    c.ConfluenceExporter.generate_toc = generate_toc
    c.ConfluenceExporter.attach_ipynb = attach_ipynb
    c.ConfluenceExporter.enable_style = enable_style
    c.ConfluenceExporter.enable_mathjax = enable_mathjax
    c.ConfluenceExporter.extra_labels = extra_labels

    exporter = ConfluenceExporter(c)
    result = exporter.from_filename(notebook_file)
    print('Updated', confluence_url)
    return result


import argparse
import getpass
import os
import sys

from .api import notebook_to_page


def main(argv=None):
    """Command line interface."""
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
        description='Converts Jupyter Notebooks to Atlassian Confluence pages using nbconvert',
        epilog="Collects credentials from the following locations:\n"
        "1. CONFLUENCE_USERNAME and CONFLUENCE_PASSWORD environment variables\n"
        "2. ~/.nbconflux file in the format username:password\n"
        "3. User prompts")
    parser.add_argument('notebook', type=str, help='Path to local notebook (ipynb)')
    parser.add_argument('url', type=str, help='URL of Confluence page to update')
    parser.add_argument('--exclude-toc', action='store_true', help='Do not generate a table of contents')
    parser.add_argument('--exclude-ipynb', action='store_true', help='Do not attach the notebook to the page')
    parser.add_argument('--exclude-style', action='store_true', help='Do not include the Jupyter base stylesheet')
    parser.add_argument('--include-mathjax', action='store_true', help='Enable MathJax on the page')
    parser.add_argument('--extra-labels', nargs='+', type=str, help='Additional labels to add to the page')

    args = parser.parse_args(argv or sys.argv[1:])

    username = os.getenv('CONFLUENCE_USERNAME')
    password = os.getenv('CONFLUENCE_PASSWORD')
    cfg = os.path.expanduser('~/.nbconflux')

    # Prefer credentials in environment variables
    if username and password:
        print('Using credentials for {} from environment variables'.format(username))
    elif os.path.isfile(cfg):
        # Fallback on credentials in a well known file location
        with open(cfg) as f:
            segs = f.read().strip().split(':', 1)
            if len(segs) == 2:
                username = segs[0]
                password = segs[1]
                print('Using credentials for {} from configuration file'.format(username))

    # Prompt the user for missing credentials
    if username is None:
        current = getpass.getuser()
        current = current[2:] if current.startswith('p-') else current
        username = input('Confluence username ({}): '.format(current))
        # Use the current username if the user doesn't enter anything
        if not username.strip():
            username = current
    if password is None:
        password = getpass.getpass('Confluence password: ')

    notebook_to_page(args.notebook, args.url, username, password,
                     generate_toc=not args.exclude_toc, attach_ipynb=not args.exclude_ipynb,
                     enable_style=not args.exclude_style, enable_mathjax=args.include_mathjax,
                     extra_labels=args.extra_labels)

if __name__ == '__main__':
    main()


from .api import notebook_to_page

# These three lines comes from versioneer.
# Add to this file if you want, but do not remove these.
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import versioneer
from setuptools import setup, find_packages

try:
    with open('README.md') as f:
        long_description = f.read()
except:
    long_description = ''
    print('Failed to load README.md as long_description')

setup(
    name='nbconflux',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Converts Jupyter Notebooks to Atlassian Confluence (R) pages using nbconvert',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Valassis Digital',
    url='https://github.com/Valassis-Digital-Media/nbconflux',
    platforms=['Linux', 'Mac OSX', 'Windows'],
    packages=find_packages(),
    include_package_data = True,
    license='BSD 3-Clause',
    entry_points = {
        'console_scripts': [
            'nbconflux = nbconflux.cli:main'
        ]
    },
    install_requires=[
        'nbconvert>=5.3'
        'requests'
        'traitlets'
    ],
)
